import numpy as np
import pandas as pd
import os
from sklearn import preprocessing
import yaml
import matplotlib.pyplot as plt
import pickle
import re
from proms import config


class Data(object):
    """Base class for data set"""

    def __init__(self, name, root, config_file, output_dir):
        self.name = name
        self.root = root
        self.all_data = None
        self.output_dir = output_dir
        if output_dir is not None and not os.path.exists(output_dir):
            os.makedirs(output_dir)
        config_file = os.path.join(root, config_file)
        with open(config_file) as fh:
            self.config = yaml.load(fh, Loader=yaml.FullLoader)
        self.has_test = self.has_test_set()

    def has_test_set(self):
        """ does the dataset contain independent test set """
        return 'test_data_directory' in self.config

    def load_data(self, data_file, samples, vis=False):
        # load feature data
        X = pd.read_csv(data_file, sep='\t', index_col=0)
        # remove features with missing values
        X.dropna(axis=0, inplace=True)

        # if there are duplicated rows (genes), take the row average
        X = X.groupby(X.index).mean()
        # sample x features
        X = X.T
        if samples is not None:
            X = X.loc[samples, :]
        if vis and self.output_dir is not None:
            myfig = plt.figure()
            X_temp = X.T
            flierprops = dict(marker='o', markersize=1)
            X_temp.boxplot(grid=False, rot=90, fontsize=2,
                flierprops=flierprops)
            upper_dir = os.path.basename(os.path.dirname(data_file))
            file_name = (self.name + '_' + upper_dir + '_' +
                         os.path.basename(data_file) + '.pdf')
            myfig.savefig(os.path.join(self.output_dir, file_name),
                          format='pdf')

        # X.sort_index(inplace=True)
        return X

    def save_data(self):
        """save all data into a file"""
        if self.output_dir is not None:
            fname = self.name + '_all_data.pkl'
            with open(os.path.join(self.output_dir, fname), 'wb') as fh:
                pickle.dump(self.all_data, fh, protocol=pickle.HIGHEST_PROTOCOL)


class Dataset(Data):
    """ data set"""
    def __init__(self, name, root, config_file, output_dir=None):
        self.clin_data = None
        self.prediction_type = None
        # only valid for classification
        self.classes = None
        super().__init__(name, root, config_file, output_dir)

    def check_prediction_type(self):
        """
        Based on the target data, infer if it is a
        regression, classification or survival analysis task
        """
        if self.clin_data is None:
            raise ValueError('No clinical data has been set')

        target_vals = self.clin_data.iloc[:,0].values
        target_dtype = target_vals.dtype
        # for classification, binary only
        if target_dtype == np.int64:
            uniq_len = len(np.unique(target_vals))
            if uniq_len == 2:
                return config.prediction_map['classification']
            if uniq_len > 2:
                return config.prediction_map['regression']
            raise ValueError('all target values are the same')

        if target_dtype == np.float64:
            return config.prediction_map['regression']

        if target_dtype == np.object:
            target_vals_str = target_vals.astype(str)
            r_surv = re.compile('.+,.+')
            if all(r_surv.match(item) for item in list(target_vals_str)):
                return config.prediction_map['survival']
            uniq_len = len(np.unique(target_vals_str))
            if uniq_len == 2:
                return config.prediction_map['classification']
            if uniq_len > 2:
                raise ValueError('multiclass classification not supported')
            raise ValueError('all target values are the same')

        raise ValueError('wrong target values')

    def load_clin_surv(self, clin_file, target_label):
        """
        load y for survival analysis
        """
        clin_data = pd.read_csv(clin_file, sep='\t', index_col=0)
        col_event = target_label[0]
        col_time = target_label[1]
        y = np.empty(dtype=[(col_event, np.bool), (col_time, np.float64)],
                     shape=clin_data.shape[0])
        y[col_event] = (clin_data[col_event] == 1).values
        y[col_time] = clin_data[col_time].values
        sample_names = clin_data.index.values
        y = pd.DataFrame.from_records(y, index=sample_names)
        return y

    def __call__(self):
        print('processing data ...')
        train_dataset = self.config['train_data_directory']
        target_view = self.config['target_view']
        target_label = self.config['target_label']
        clin_file = self.config['data']['train']['label']['file']
        clin_file = os.path.join(self.root, train_dataset, clin_file)
        if isinstance(target_label, list):  # survival
            clin_data = self.load_clin_surv(clin_file, target_label)
            train_sample = clin_data.index
            self.clin_data = clin_data
            self.prediction_type = config.prediction_map['survival']
        else:
            # the first col will be the index
            clin_data = pd.read_csv(clin_file, sep='\t', index_col=0)
            # samples X phenotype
            clin_data = clin_data.loc[:, [target_label]]
            self.clin_data = clin_data
            self.prediction_type = self.check_prediction_type()
            train_sample = clin_data.index
        print(f'prediction type: {self.prediction_type}')
        # for now: must be two-class classification
        if self.prediction_type == config.prediction_map['classification']:
            le = preprocessing.LabelEncoder()
            clin_vals = clin_data.iloc[:, 0].values
            le.fit(clin_vals)
            self.classes = le.classes_
            clin_data.iloc[:, 0] = le.transform(clin_vals)
            self.clin_data = clin_data

        y = clin_data

        # get the name of samples that have all the omics data available
        all_views = self.config['data']['train']['view']
        n_views = len(all_views)
        all_view_names = [item['type'] for item in all_views]
        all_view_names.remove(target_view)
        all_view_ordered = [target_view] + sorted(all_view_names)

        if self.has_test:
            test_dataset = self.config['test_data_directory']
            all_test_views = self.config['data']['test']['view']
            if 'label' in self.config['data']['test']:
                test_clin_file = self.config['data']['test']['label']['file']
                test_clin_file = os.path.join(self.root, test_dataset, test_clin_file)
                if self.prediction_type == 'sur':
                    test_clin_data = self.load_clin_surv( test_clin_file, target_label)
                    test_samples = test_clin_data.index
                else:
                    test_clin_data = pd.read_csv(test_clin_file, sep='\t',
                                                 index_col=0)
                    test_samples = test_clin_data.index
                    # samples X phenotype
                    test_clin_data = test_clin_data.loc[:, [target_label]]
                if self.prediction_type == config.prediction_map['classification']:
                    le = preprocessing.LabelEncoder()
                    test_clin_vals = test_clin_data.iloc[:, 0].values
                    le.fit(test_clin_vals)
                    if any(le.classes_ != self.classes):
                        raise ValueError('class label in test dataset not'
                            ' matching training data')
                    test_clin_data.iloc[:, 0] = le.transform(test_clin_vals)
                y_final_test_2 = test_clin_data
            else:
                # no label provided, only for prediction
                test_samples = None
                y_final_test_2 = None
            selected_view = filter(lambda view: view['type'] == target_view,
                                   all_test_views)
            test_view_file = list(selected_view)[0]['file']
            test_view_file = os.path.join(self.root, test_dataset, test_view_file)
            X_final_test_2 = self.load_data(test_view_file, test_samples)
        else:
            X_final_test_2 = None
            y_final_test_2 = None

        all_data_ = {'desc': {
                        'prediction_type': self.prediction_type,
                        'name': self.name,
                        'has_test': self.has_test,
                        'target_label': target_label,
                        'target_view': target_view,
                        'n_view': n_views,
                        'view_names': all_view_ordered
                        },
                     'X': {},
                     'y': y,
                     'X_test': None,
                     'y_test': y_final_test_2
                    }

        for i in range(n_views):
            view_name = all_views[i]['type']
            print(f'view: {view_name}')
            view_file = os.path.join(self.root, train_dataset,
                                     all_views[i]['file'])
            X = self.load_data(view_file, train_sample)
            if view_name == target_view:
                if self.has_test:
                    # find out common features in the two data
                    # sets for the target view
                    common_features = sorted(list(set(X.columns.values) & set(
                        X_final_test_2.columns.values)))
                    print('train data target view features length {}'.format(
                        len(X.columns.values)))
                    print('test data target view features length {}'.format(
                        len(X_final_test_2.columns.values)))
                    print('common target view features length {}'.format(
                        len(common_features)))
                    X = X[common_features]
                    X_final_test_2 = X_final_test_2[common_features]
                    X_final_test_2 = X_final_test_2.add_prefix(view_name + '_')

            X = X.add_prefix(view_name + '_')
            all_data_['X'][view_name] = X

        if self.has_test:
            all_data_['X_test'] = X_final_test_2

        print('number of samples: {}'.format(all_data_['y'].shape[0]))

        if self.has_test:
            all_data_['X_test'] = X_final_test_2
            if all_data_['y_test'] is not None:
                print('test sample size: {}'.format(all_data_['y_test'].shape[0]))
            else:
                print('test sample size: {}'.format(all_data_['X_test'].shape[0]))

        self.all_data = all_data_
        # self.save_data()
        return all_data_



"""K-medoids clustering"""
import warnings
import numpy as np
from sklearn.base import BaseEstimator, ClusterMixin, TransformerMixin
from sklearn.metrics.pairwise import pairwise_distances
from sklearn.metrics.pairwise import pairwise_distances_argmin
from sklearn.utils import check_array, check_random_state
from sklearn.utils.extmath import stable_cumsum
from sklearn.utils.validation import check_is_fitted, _check_sample_weight
from sklearn.utils.validation import _num_samples
from sklearn.exceptions import ConvergenceWarning
from scipy.spatial import distance
from scipy.spatial.distance import _validate_vector
from scipy.linalg import norm


def abs_correlation(u, v):
    """
    absolute pearson correlation distance (1-abs(pearson_cc))
    """
    u = _validate_vector(u)
    v = _validate_vector(v)
    umu = u.mean()
    vmu = v.mean()
    um = u - umu
    vm = v - vmu
    dist = 1.0 - np.abs(np.dot(um, vm) / (norm(um) * norm(vm)))
    return dist


class KMedoids(BaseEstimator, ClusterMixin, TransformerMixin):
    """weighted k-medoids clustering.

    Read more in the :ref:`User Guide <k_medoids>`.

    Parameters
    ----------
    n_clusters : int, optional, default: 8
        The number of clusters to form as well as the number of medoids to
        generate.

    metric : string, or callable, optional, default: 'euclidean'
        What distance metric to use. See :func:metrics.pairwise_distances

    init : {'random', 'heuristic', 'k-medoids++'}, optional,
        default: 'heuristic'
        Specify medoid initialization method. 'random' selects n_clusters
        elements from the dataset. 'heuristic' picks the n_clusters points
        with the smallest sum distance to every other point. 'k-medoids++'
        follows an approach based on k-means++_, and in general, gives initial
        medoids which are more separated than those generated by the other
        methods.

        .. _k-means++:
        https://theory.stanford.edu/~sergei/papers/kMeansPP-soda.pdf

    max_iter : int, optional, default : 300
        Specify the maximum number of iterations when fitting.

    random_state : int, RandomState instance or None, optional
        Specify random state for the random number generator. Used to
        initialise medoids when init='random'.

    w: array, shape = (n_samples,), optional, default: None
       Specify the sample weights

    Attributes
    ----------
    cluster_centers_ : array, shape = (n_clusters, n_features)
            or None if metric == 'precomputed'
        Cluster centers, i.e. medoids (elements from the original dataset)

    medoid_indices_ : array, shape = (n_clusters,)
        The indices of the medoid rows in X

    labels_ : array, shape = (n_samples,)
        Labels of each point

    inertia_ : float
        Sum of distances of samples to their closest cluster center.

    Examples
    --------
    >>> import numpy as np

    >>> X = np.asarray([[1, 2], [1, 4], [1, 0],
    ...                 [4, 2], [4, 4], [4, 0]])
    >>> kmedoids = KMedoids(n_clusters=2, random_state=0).fit(X)
    >>> kmedoids.labels_
    array([0, 0, 0, 1, 1, 1])
    >>> kmedoids.predict([[0,0], [4,4]])
    array([0, 1])
    >>> kmedoids.cluster_centers_
    array([[1, 2],
           [4, 2]])
    >>> kmedoids.inertia_
    8.0

    See scikit-learn-extra/examples/plot_kmedoids_digits.py for examples
    of KMedoids with various distance metrics.

    References
    ----------
    Kaufman, L. and Rousseeuw, P.J., Statistical Data Analysis Based on
    the L1–Norm and Related Methods, edited by Y. Dodge, North-Holland,
    405–416. 1987

    See also
    --------

    KMeans
        The KMeans algorithm minimizes the within-cluster sum-of-squares
        criterion. It scales well to large number of samples.

    Notes
    -----
    Since all pairwise distances are calculated and stored in memory for
    the duration of fit, the space complexity is O(n_samples ** 2).

    """

    def __init__(
        self,
        n_clusters=8,
        metric="euclidean",
        init="heuristic",
        max_iter=300,
        n_jobs=None,
        random_state=None,
    ):
        self.n_clusters = n_clusters
        self.metric = metric
        self.init = init
        self.max_iter = max_iter
        self.n_jobs = n_jobs
        self.random_state = random_state

    def _check_nonnegative_int(self, value, desc):
        """Validates if value is a valid integer > 0"""

        if (
            value is None
            or value <= 0
            or not isinstance(value, (int, np.integer))
        ):
            raise ValueError(
                "%s should be a nonnegative integer. "
                "%s was given" % (desc, value)
            )

    def _check_normalized_sample_weight(self, sample_weight, X):
        """Set sample_weight to 1 if None
           if provided, normalize them to within the range of (0,1)
        """

        sample_weight_was_none = sample_weight is None
        sample_weight = _check_sample_weight(sample_weight, X, dtype=X.dtype)
        # require postive weight only
        assert np.all(sample_weight > 0), 'weight must be positive'
        # if weights are in [0, 1], do nothing
        if not sample_weight_was_none:
            all_in_range = ((sample_weight >= 0).all() and
                            (sample_weight <= 1).all())
            if not all_in_range:
                # 2*(1-sigmoid(x)) -> (0,1), normalized to between 0 and 1
                sample_weight = 2.0 * (1 / (1 + np.exp(-sample_weight)) - 0.5)
        return sample_weight

    def _check_candidacy(self, candidacy, X):
        """"check candidacy array values"""

        n_samples = _num_samples(X)

        if candidacy is None:
            candidacy = np.ones(n_samples, dtype=bool)
        else:
            if candidacy.ndim != 1:
                raise ValueError('candidacy must be 1D array')

            if candidacy.shape != (n_samples,):
                raise ValueError('candidacy.shape == {}, expected {}!'
                                 .format(candidacy.shape, (n_samples,)))

            if np.issubdtype(candidacy.dtype, np.integer):
                candidacy = candidacy.astype(np.bool)
            elif candidacy.dtype != np.bool:
                raise TypeError('candidacy type must be integer or bool')

        return candidacy

    def _check_init_args(self):
        """Validates the input arguments. """

        # Check n_clusters and max_iter
        self._check_nonnegative_int(self.n_clusters, "n_clusters")
        self._check_nonnegative_int(self.max_iter, "max_iter")

        # Check init
        init_methods = ["random", "heuristic", "k-medoids++"]
        if self.init not in init_methods:
            raise ValueError(
                "init needs to be one of "
                + "the following: "
                + "%s" % init_methods
            )

    def fit(self, X, y=None, sample_weight=None, candidacy=None):
        """Fit K-Medoids to the provided data.

        Parameters
        ----------
        X : {array-like, sparse matrix}, shape = (n_samples, n_features), \
                or (n_samples, n_samples) if metric == 'precomputed'
            Dataset to cluster.

        y : Ignored

        sample_weight : array-like, shape (n_samples,), optional
            The weights for each observation in X. If None, all observations
            are assigned equal weight (default: None).

        candidacy:  array-like, shape (n_sample,), optional
           Only the samples indicated as True in the candidacy can be selected
           as cluster center, if candidacy is None,  any sample can be
           potentially selected as center.

        Returns
        -------
        self
        """
        random_state_ = check_random_state(self.random_state)

        self._check_init_args()
        X = check_array(X, accept_sparse=["csr", "csc"])
        if self.n_clusters > X.shape[0]:
            warn_msg = f'The number of medoids ({self.n_clusters}) must be less ' \
                       f'than the number of samples ({X.shape[0]})'
            warnings.warn(warn_msg)

        cnsw = self._check_normalized_sample_weight
        checked_sample_weight = cnsw(sample_weight, X)
        checked_candidacy = self._check_candidacy(candidacy, X)

        n_cand = np.sum(checked_candidacy)
        if n_cand < self.n_clusters:
            warn_msg = f'The number of candidancy {n_cand} cannot be less '   \
                       f'than the number of clusters {self.n_clusters}'
            warnings.warn(warn_msg)

        D = pairwise_distances(X, metric=self.metric, n_jobs=self.n_jobs)
        medoid_idxs = self._initialize_medoids(D, self.n_clusters,
                                               random_state_,
                                               checked_candidacy)
        labels = None

        # Continue the algorithm as long as
        # the medoids keep changing and the maximum number
        # of iterations is not exceeded
        for self.n_iter_ in range(0, self.max_iter):
            old_medoid_idxs = np.copy(medoid_idxs)
            labels = np.argmin(D[medoid_idxs, :], axis=0)

            # Update medoids with the new cluster indices
            self._update_medoid_idxs_in_place(D, labels, medoid_idxs,
                                              checked_sample_weight,
                                              checked_candidacy)
            if np.all(old_medoid_idxs == medoid_idxs):
                break
            elif self.n_iter_ == self.max_iter - 1:
                warnings.warn(
                    "Maximum number of iteration reached before "
                    "convergence. Consider increasing max_iter to "
                    "improve the fit.",
                    ConvergenceWarning,
                )

        # Set the resulting instance variables.
        if self.metric == "precomputed":
            self.cluster_centers_ = None
        else:
            self.cluster_centers_ = X[medoid_idxs]

        # Expose labels_ which are the assignments of
        # the training data to clusters
        self.labels_ = labels
        self.medoid_indices_ = medoid_idxs
        self.inertia_ = self._compute_inertia(self.transform(X))

        # Return self to enable method chaining
        return self

    def _update_medoid_idxs_in_place(self, D, labels, medoid_idxs,
                                     checked_sample_weight, checked_candidacy):
        """In-place update of the medoid indices"""

        # Update the medoids for each cluster
        for k in range(self.n_clusters):
            # Extract the distance matrix between the data points
            # inside the cluster k
            cluster_k_idxs = np.where(labels == k)[0]

            if len(cluster_k_idxs) == 0:
                warnings.warn(
                    "Cluster {k} is empty! "
                    "self.labels_[self.medoid_indices_[{k}]] "
                    "may not be labeled with "
                    "its corresponding cluster ({k}).".format(k=k)
                )
                continue

            in_cluster_distances = D[np.ix_(cluster_k_idxs, cluster_k_idxs)]
            in_cluster_candidacy = checked_candidacy[cluster_k_idxs]

            # Calculate all costs from each point to all others in
            # the cluster (weighted)
            in_cluster_w = checked_sample_weight[cluster_k_idxs]
            in_cluster_all_costs = in_cluster_distances.dot(in_cluster_w)

            # select the one with min cost among all candidates
            order = np.argsort(in_cluster_all_costs)
            min_cost_idx = order[in_cluster_candidacy[order]][0]
            min_cost = in_cluster_all_costs[min_cost_idx]
            curr_cost = in_cluster_all_costs[np.argmax(cluster_k_idxs ==
                                                       medoid_idxs[k])]

            # Adopt a new medoid if its distance is smaller than the current
            if min_cost < curr_cost:
                medoid_idxs[k] = cluster_k_idxs[min_cost_idx]

    def fit_predict(self, X, y=None, sample_weight=None, candidacy=None):
        """Compute cluster centers and predict cluster index for each sample.

        Convenience method; equivalent to calling fit(X) followed by
        predict(X).

        Parameters
        ----------
        X : {array-like, sparse matrix} of shape (n_samples, n_features)
            New data to transform.

        y : Ignored
            Not used, present here for API consistency by convention.

        sample_weight : array-like, shape (n_samples,), optional
            The weights for each observation in X. If None, all observations
            are assigned equal weight (default: None).

        candidacy:  array-like, shape (n_sample,), optional
           Only the samples indicated as True in the candidacy can be selected
           as cluster center, if candidacy is None,  any sample can be
           potentially
           selected as center.

        Returns
        -------
        labels : array, shape [n_samples,]
            Index of the cluster each sample belongs to.
        """
        return self.fit(X, sample_weight=sample_weight,
                        candidacy=candidacy).labels_

    def transform(self, X):
        """Transforms X to cluster-distance space.

        Parameters
        ----------
        X : {array-like, sparse matrix}, shape (n_query, n_features), \
                or (n_query, n_indexed) if metric == 'precomputed'
            Data to transform.

        Returns
        -------
        X_new : {array-like, sparse matrix}, shape=(n_query, n_clusters)
            X transformed in the new space of distances to cluster centers.
        """
        X = check_array(X, accept_sparse=["csr", "csc"])

        if self.metric == "precomputed":
            check_is_fitted(self, "medoid_indices_")
            return X[:, self.medoid_indices_]
        else:
            check_is_fitted(self, "cluster_centers_")

            Y = self.cluster_centers_
            return pairwise_distances(X, Y=Y, metric=self.metric,
                                      n_jobs=self.n_jobs)

    def predict(self, X):
        """Predict the closest cluster for each sample in X.

        Parameters
        ----------
        X : {array-like, sparse matrix}, shape (n_query, n_features), \
                or (n_query, n_indexed) if metric == 'precomputed'
            New data to predict.

        Returns
        -------
        labels : array, shape = (n_query,)
            Index of the cluster each sample belongs to.
        """
        X = check_array(X, accept_sparse=["csr", "csc"])

        if self.metric == "precomputed":
            check_is_fitted(self, "medoid_indices_")
            return np.argmin(X[:, self.medoid_indices_], axis=1)
        else:
            check_is_fitted(self, "cluster_centers_")

            # Return data points to clusters based on which cluster assignment
            # yields the smallest distance
            return pairwise_distances_argmin(
                X, Y=self.cluster_centers_, metric=self.metric
            )

    def _compute_inertia(self, distances):
        """Compute inertia of new samples. Inertia is defined as the sum of the
        sample distances to closest cluster centers.

        Parameters
        ----------
        distances : {array-like, sparse matrix}, shape=(n_samples, n_clusters)
            Distances to cluster centers.

        Returns
        -------
        Sum of sample distances to closest cluster centers.
        """

        # Define inertia as the sum of the sample-distances
        # to closest cluster centers
        inertia = np.sum(np.min(distances, axis=1))

        return inertia

    def _initialize_medoids(self, D, n_clusters, random_state_,
                            checked_candidacy):
        """Select initial mediods when beginning clustering."""

        if self.init == "random":  # Random initialization
            # Pick random k medoids as the initial ones from candidacies
            candidacy = np.where(checked_candidacy)[0]
            medoids = random_state_.choice(candidacy, n_clusters)
        elif self.init == "k-medoids++":
            medoids = self._kpp_init(D, n_clusters, random_state_,
                                     checked_candidacy)
        elif self.init == "heuristic":  # Initialization by heuristic
            # Pick K first data points that have the smallest sum distance
            # to every other point. These are the initial medoids.
            # sort first
            order = np.argsort(np.sum(D, axis=1))
            medoids = order[checked_candidacy[order]][:n_clusters]
        else:
            raise ValueError(
                "init value '{init}' not recognized".format(init=self.init)
            )

        return medoids

    # Copied from sklearn.cluster.k_means_._k_init
    def _kpp_init(self, D, n_clusters, random_state_, checked_candidacy,
                  n_local_trials=None):
        """Init n_clusters seeds with a method similar to k-means++

        Parameters
        -----------
        D : array, shape (n_samples, n_samples)
            The distance matrix we will use to select medoid indices.

        n_clusters : integer
            The number of seeds to choose

        random_state : RandomState
            The generator used to initialize the centers.

        checked_candidacy:

        n_local_trials : integer, optional
            The number of seeding trials for each center (except the first),
            of which the one reducing inertia the most is greedily chosen.
            Set to None to make the number of trials depend logarithmically
            on the number of seeds (2+log(k)); this is the default.

        Notes
        -----
        Selects initial cluster centers for k-medoid clustering in a smart way
        to speed up convergence. see: Arthur, D. and Vassilvitskii, S.
        "k-means++: the advantages of careful seeding". ACM-SIAM symposium
        on Discrete algorithms. 2007

        Version ported from http://www.stanford.edu/~darthur/kMeansppTest.zip,
        which is the implementation used in the aforementioned paper.
        """
        n_samples, _ = D.shape

        centers = np.empty(n_clusters, dtype=int)

        # Set the number of local seeding trials if none is given
        if n_local_trials is None:
            # This is what Arthur/Vassilvitskii tried, but did not report
            # specific results for other than mentioning in the conclusion
            # that it helped.
            n_local_trials = 2 + int(np.log(n_clusters))

        # select first center from candidates
        center_id = random_state_.choice(np.where(checked_candidacy)[0], 1)
        centers[0] = center_id

        # Initialize list of closest distances and calculate current potential
        closest_dist_sq = D[centers[0], :] ** 2
        current_pot = closest_dist_sq.sum()

        # pick the remaining n_clusters-1 points
        for cluster_index in range(1, n_clusters):
            # Choose center candidates by sampling with prob. proportional
            # to the squared distance to the closest existing center
            candidate_ids = []
            # possibly never ending ??
            while len(candidate_ids) != n_local_trials:
                rand_val = (random_state_.random_sample(1) * current_pot)
                candidate_id = np.searchsorted(stable_cumsum(closest_dist_sq),
                                               rand_val)
                # XXX: numerical imprecision can result in a candidate_id out
                # of range
                candidate_id = np.clip(candidate_id, None,
                                       closest_dist_sq.size - 1)
                if checked_candidacy[candidate_id]:
                    candidate_ids.append(candidate_id)

            assert len(candidate_ids) == n_local_trials

            # Compute distances to center candidates
            distance_to_candidates = D[candidate_ids, :] ** 2

            # Decide which candidate is the best
            best_candidate = None
            best_pot = None
            best_dist_sq = None
            for trial in range(n_local_trials):
                # Compute potential when including center candidate
                new_dist_sq = np.minimum(
                    closest_dist_sq, distance_to_candidates[trial]
                )
                new_pot = new_dist_sq.sum()

                # Store result if it is the best local trial so far
                if (best_candidate is None) or (new_pot < best_pot):
                    best_candidate = candidate_ids[trial]
                    best_pot = new_pot
                    best_dist_sq = new_dist_sq

            centers[cluster_index] = best_candidate
            current_pot = best_pot
            closest_dist_sq = best_dist_sq

        return centers



from .fs_methods import ProMS, ProMS_mo
from .fs_methods import fs_methods, FeatureSelector

from .dataset import Dataset, Data
__all__ = [
           'Data',
           'Dataset',
           'ProMS',
           'ProMS_mo',
           'fs_methods',
           'FeatureSelector'
           ]



import numpy as np
from scipy.sparse.construct import rand
from xgboost import XGBClassifier, XGBRegressor
from sklearn.ensemble import RandomForestClassifier, RandomForestRegressor
from sklearn.svm import SVC, SVR
from sklearn.linear_model import LogisticRegression
from sksurv.linear_model import CoxPHSurvivalAnalysis
from sksurv.ensemble import RandomSurvivalForest
from sksurv.ensemble import GradientBoostingSurvivalAnalysis
from sksurv.svm import FastSurvivalSVM
from sklearn.linear_model import Ridge


"""
define package wise constants
"""

prediction_map = {
      'classification': 'cls',
      'regression': 'reg',
      'survival': 'sur'
  }


default_estimators = {
  'cls': ['lr', 'rf', 'svm', 'gbm'],
  'reg': ['ridge', 'rf', 'svm', 'gbm'],
#   'sur': ['coxph', 'rf', 'svm', 'gbm']
  'sur': ['coxph']
}

parameter_grid = {
    'cls': {
        'lr': {
            'C': [0.001, 0.01, 0.1, 1, 10, 100]
        },
        'rf':{
            'max_depth': [2, 4, 6, 8, 10],
            'n_estimators': range(50, 400, 50)
        },
        'svm':{
            'C': [0.001, 0.01, 0.1, 1, 10, 100],
            'gamma': [0.001, 0.01, 0.1, 1, 10, 100]
        },
        'gbm':{
            'max_depth': [2, 4, 6, 8, 10],
            'n_estimators': range(50, 400, 50)
        }
    },
    'reg': {
        'ridge':{
            'alpha': 10. ** np.linspace(-4, 4, 10)
        },
        'rf':{
            'max_depth': [2, 4, 6, 8, 10],
            'n_estimators': range(50, 400, 50)
        },
        'svm':{
            'C': [0.001, 0.01, 0.1, 1, 10, 100],
            'gamma': [0.001, 0.01, 0.1, 1, 10, 100]
        },
        'gbm':{
            'max_depth': [2, 4, 6, 8, 10],
            'n_estimators': range(50, 400, 50)
        }
    },
    'sur': {
        'svm': {
            'alpha': 10. ** np.linspace(-4, 4, 10)
        },
        'coxph': {
            'alpha': 10. ** np.linspace(-4, 4, 10)
        },
        'gbm': {
            'max_depth': [2, 4, 6, 8, 10],
            'n_estimators': range(50, 400, 50)
        },
        'rf': {
            'max_depth': [2, 4, 6, 8, 10],
            'n_estimators': range(50, 400, 50)
        }
    }
}


def get_estimator(seed, estimator, prediction_type='cls'):
    """
    classfication:
        lr: sklearn.linear_model.LogisticRegression
        rf: sklearn.ensemble.RandomForestClassifier
        svm: sklearn.svm.SVC
        gbm: xgboost.XGBClassifier
    regression:
        rf: sklearn.ensemble.RandomForestRegressor
        svm: sklearn.svm.SVR
        gbm xgboost.XGBRegressor
    survival:
        coxph: sksurv.linear_model.CoxnetSurvivalAnalysis
        rf: sksurv.ensemble.RandomSurvivalForest
        svm: sksurv.svm.FastKernelSurvivalSVM
        gbm: sksurv.ensemble.GradientBoostingSurvivalAnalysis
    """
    if not estimator in default_estimators[prediction_type]:
        raise ValueError(f'estimator {estimator} not supported for prediction type {prediction_type}')

    if prediction_type == 'cls':
        est_dict = {
            'lr': LogisticRegression(max_iter=1000, random_state=seed),
            'rf': RandomForestClassifier(random_state=seed),
            'svm': SVC(kernel='rbf', probability=True, random_state=seed),
            'gbm': XGBClassifier(random_state=seed)
        }
        return est_dict[estimator]

    if prediction_type == 'reg':
        est_dict = {
            'ridge': Ridge(random_state=seed),
            'rf': RandomForestRegressor(random_state=seed),
            'svm': SVR(kernel='rbf'),
            'gbm': XGBRegressor(random_state=seed)
        }
        return est_dict[estimator]

    if prediction_type == 'sur':
        est_dict = {
            'coxph': CoxPHSurvivalAnalysis(),
            # 'rf': RandomSurvivalForest(random_state=seed),
            # 'svm': FastSurvivalSVM(random_state=seed),
            # 'gbm': GradientBoostingSurvivalAnalysis(random_state=seed)
        }
        return est_dict[estimator]
    



import multiprocessing
import warnings
import os
import json
import yaml
import csv
from datetime import datetime
from tempfile import mkdtemp
from shutil import rmtree
import pandas as pd
import numpy as np
import argparse
import pickle
from sklearn.model_selection import train_test_split
from sklearn.metrics import accuracy_score, mean_squared_error
from sklearn.model_selection import GridSearchCV
from sklearn.model_selection import StratifiedKFold, KFold
from sklearn.pipeline import Pipeline
# from .utils import StandardScalerDf
from joblib import Memory
from proms import Dataset, FeatureSelector, config


def check_prediction_type_ok(prediction_type):
    return True if prediction_type in ['cls', 'reg', 'sur'] else False


def get_scores(prediction_type, grid, X_test, y_test, X_test_2, y_test_2):
    """
     evaluate on the test data set (if available)
     and independent test set (if available)
    """
    scores = None
    if prediction_type == 'cls':
        test_acc = test_score = 'NA'
        label = pred_label_s = pred_prob_s = 'NA'
        test_acc_2 = test_score_2 = 'NA'
        label_2 = pred_label_s_2 = pred_prob_s_2 = 'NA'
        if X_test is not None:
            label = ','.join(map(str, y_test))
            pred_prob = grid.predict_proba(X_test)[:, 1]
            pred_label = [1 if x >= 0.5 else 0 for x in pred_prob]
            test_score = grid.score(X_test, y_test)
            pred_prob_s = ','.join(map('{:.4f}'.format, pred_prob))
            pred_label_s = ','.join(map(str, pred_label))
            test_acc = accuracy_score(y_test, pred_label)

        if X_test_2 is not None:
            pred_prob_2 = grid.predict_proba(X_test_2)[:, 1]
            pred_label_2 = [1 if x >= 0.5 else 0 for x in pred_prob_2]
            pred_prob_s_2 = ','.join(map('{:.4f}'.format, pred_prob_2))
            pred_label_s_2 = ','.join(map(str, pred_label_2))
            if y_test_2 is not None:
                label_2 = ','.join(map(str, y_test_2))
                test_score_2 = round(grid.score(
                    X_test_2, y_test_2), 4)
                test_acc_2 = round(accuracy_score(y_test_2, pred_label_2), 4)
        scores = {'test_acc': test_acc,
                  'test_score': test_score,
                  'pred_prob_s': pred_prob_s,
                  'pred_label_s': pred_label_s,
                  'label': label,
                  'pred_prob_s_2': pred_prob_s_2,
                  'pred_label_s_2': pred_label_s_2,
                  'label_2': label_2,
                  'test_acc_2': test_acc_2,
                  'test_score_2': test_score_2
                 }
    elif prediction_type == 'reg':
        test_mse = test_score = 'NA'
        label = pred_label_s = 'NA'
        test_mse_2 = test_score_2 = 'NA'
        label_2 = pred_label_s_2 = 'NA'
        if X_test is not None:
            label = ','.join(map(str, y_test))
            pred_label = grid.predict(X_test)
            pred_label = [round(item, 4) for item in pred_label]
            test_score = grid.score(X_test, y_test)
            pred_label_s = ','.join(map(str, pred_label))
            test_mse = mean_squared_error(y_test, pred_label)

        if X_test_2 is not None:
            pred_label_2 = grid.predict(X_test_2)
            pred_label_s_2 = ','.join(map(str, pred_label_2))
            if y_test_2 is not None:
                label_2 = ','.join(map(str, y_test_2))
                test_score_2 = round(grid.score(X_test_2, y_test_2), 4)
                test_mse_2 = round(mean_squared_error(y_test_2, pred_label_2), 4)
        scores = {'test_mse': test_mse,
                  'test_score': test_score,
                  'pred_label_s': pred_label_s,
                  'label': label,
                  'pred_label_s_2': pred_label_s_2,
                  'label_2': label_2,
                  'test_mse_2': test_mse_2,
                  'test_score_2': test_score_2
                 }
    elif prediction_type == 'sur':
        test_risk_score_s = test_risk_score_s_2 = 'NA'
        test_c_index = test_c_index_2 = 'NA'
        label = label_2 = 'NA'
        if X_test is not None:
            label = ','.join(map(str, y_test))
            test_risk_score = grid.predict(X_test)
            test_c_index = round(grid.score(X_test, y_test), 4)
            test_risk_score_s = ','.join(map('{:.4f}'.format, test_risk_score))

        if X_test_2 is not None:
            test_risk_score_2 = grid.predict(X_test_2)
            test_risk_score_s_2 = ','.join(map('{:.4f}'.format, test_risk_score_2))
            if y_test_2 is not None:
                label_2 = ','.join(map(str, y_test_2))
                test_c_index_2 = round(grid.score(X_test_2, y_test_2), 4)
        scores = {'test_risk_score': test_risk_score_s,
                  'label': label,
                  'test_score': test_c_index,
                  'test_risk_score_2': test_risk_score_s_2,
                  'label_2': label_2,
                  'test_score_2': test_c_index_2
                 }
    
    return scores


def set_up_results(data, mode, run_config, prediction_type, fs_method,
                   k, estimator, repeat, scores, grid):
    best_fs = grid.best_estimator_.named_steps['featureselector']
    selected_features = best_fs.get_feature_names()
    res = None

    if mode == 'full':
        cluster_membership = best_fs.get_cluster_membership()
        run_version = run_config['run_version']
        output_root = run_config['output_root']
        out_dir_run = os.path.join(output_root, run_version)
        out_dir_run_full = os.path.join(out_dir_run, 'full_model')
        with open(os.path.join(out_dir_run_full, 'full_model.pkl'), 'wb') as fh:
            pickle.dump(grid.best_estimator_, fh,
                        protocol=pickle.HIGHEST_PROTOCOL)

    omics_type = 'so' if fs_method == 'proms' else 'mo'
    if mode == 'eval':
        if prediction_type == 'cls':
            test_acc_1 = round(scores['test_acc'], 4)
            test_score_1 = round(scores['test_score'], 4)
            res = [fs_method, omics_type, k,
                   estimator, repeat, scores['pred_prob_s'],
                   scores['pred_label_s'], scores['label'],
                   test_acc_1, test_score_1]
            print(f'acc:{test_acc_1}, auroc: {test_score_1}')
        elif prediction_type == 'reg':
            test_mse_1 = round(scores['test_mse'], 4)
            test_score_1 = round(scores['test_score'], 4)
            res = [fs_method, omics_type, k,
                   estimator, repeat, scores['pred_label_s'], scores['label'],
                   test_mse_1, test_score_1]
            print(f'mse:{test_mse_1}, r2: {test_score_1}')
        elif prediction_type == 'sur':
            test_score_1 = round(scores['test_score'], 4)
            res = [fs_method, omics_type, k,
                   estimator, repeat, scores['label'], 
                   scores['test_risk_score'], test_score_1]
            print(f'c-index: {test_score_1}')
    elif mode == 'full':
        if fs_method != 'pca_ex':
            s_features = ','.join(selected_features)
        else:
            s_features = 'NA'
        if data['test_2'] is not None and data['test_2']['X'] is not None:
            if data['test_2']['y'] is not None:
                if prediction_type == 'cls':
                    test_acc_2 = round(scores['test_acc_2'], 4)
                    test_score_2 = round(scores['test_score_2'], 4)
                    res = [fs_method, omics_type, k, estimator,
                           s_features, json.dumps(cluster_membership), -1, -1,
                           scores['pred_prob_s_2'], scores['pred_label_s_2'],
                           scores['label_2'], test_acc_2, test_score_2]
                    print(f'acc:{test_acc_2}, auroc:{test_score_2}')
                elif prediction_type == 'reg':
                    test_mse_2 = round(scores['test_mse_2'], 4)
                    test_score_2 = round(scores['test_score_2'], 4)
                    res = [fs_method, omics_type, k, estimator,
                           s_features, json.dumps(cluster_membership), -1, -1,
                           scores['pred_label_s_2'],
                           scores['label_2'], test_mse_2, test_score_2]
                    print(f'mse:{test_mse_2}, r2:{test_score_2}')
                elif prediction_type == 'sur':
                    test_score_2 = round(scores['test_score_2'], 4)
                    res = [fs_method, omics_type, k, estimator,
                           s_features, json.dumps(cluster_membership), -1,
                           scores['label_2'], 
                           scores['test_risk_score_2'], test_score_2]
                    print(f'c-index:{test_score_2}')
            else:
                if prediction_type == 'cls':
                    res = [fs_method, omics_type, k, estimator,
                        s_features, json.dumps(cluster_membership), -1, -1,
                        scores['pred_prob_s_2'], scores['pred_label_s_2']]
                elif prediction_type == 'reg':
                    res = [fs_method, omics_type, k, estimator,
                        s_features, json.dumps(cluster_membership), -1, -1,
                        scores['pred_label_s_2']]
                elif prediction_type == 'sur':
                    res = [fs_method, omics_type, k, estimator,
                        s_features, json.dumps(cluster_membership), -1,
                        scores['test_risk_score_2']]
        else:
            if prediction_type == 'sur':
                res = [fs_method, omics_type, k, estimator,
                    s_features, json.dumps(cluster_membership), -1]
            else:
                res = [fs_method, omics_type, k, estimator,
                    s_features, json.dumps(cluster_membership), -1, -1]

    return res


def run_single_fs_method(data, fs_method, run_config, k, repeat, estimator,
                         mode, seed):
    n_jobs = run_config['n_jobs']
    percentile = run_config['percentile']
    prediction_type = data['desc']['prediction_type']
    est = config.get_estimator(seed, estimator, prediction_type)

    print(f'k={k}, repeat={repeat+1}, estimator={estimator},'
          f' fs_method={fs_method}', flush=True)

    # pipeline steps
    target_view_name = data['desc']['target_view_name']
    X_train_combined = None
    y_train = data['train']['y']
    X_test_combined = y_test = None
    X_test_2 = y_test_2 =  None
    view_names = data['desc']['view_names']
    for view in view_names:
        cur_train = data['train'][view]
        X_train_combined = pd.concat([X_train_combined, cur_train['X']],
                                     axis=1)
        if data['test'] is not None:
            cur_test = data['test'][view]
            X_test_combined = pd.concat([X_test_combined, cur_test['X']],
                                        axis=1)
    if data['test'] is not None:
        y_test = data['test']['y']
        y_test = get_y(y_test, prediction_type)
    if data['test_2'] is not None:
        X_test_2 = data['test_2']['X']
        y_test_2 = data['test_2']['y']
        y_test_2 = get_y(y_test_2, prediction_type)

    p_steps = []
    fs = FeatureSelector(views=view_names,
                         target_view=target_view_name,
                         method=fs_method,
                         k=k,
                         weighted=True,
                         prediction_type=prediction_type)
    # assume all data are standardized
    # it is a bit tricky to add this to the pipeline
    # because the independent test data may have
    # different number of input features, so the full model trained
    # can not be directly used to make predictions
    p_steps.append(('featureselector', fs))
    p_steps.append((estimator, est))
    cachedir = mkdtemp()
    memory = Memory(location=cachedir, verbose=0)
    pipe = Pipeline(steps=p_steps, memory=memory)
    param_grid = get_estimator_parameter_grid(estimator, prediction_type)
    param_grid['featureselector__percentile'] = percentile
    score_func = {
        'cls': 'roc_auc',
        'reg': 'r2',
         # use the default scorer of the estimator
        'sur': None
    }
    cv = {
        'cls': StratifiedKFold(3),
        'reg': KFold(3),
        'sur': KFold(3)
    }
    grid = GridSearchCV(pipe, param_grid, scoring=score_func[prediction_type],
                        cv=cv[prediction_type],
                        n_jobs=n_jobs,
                        verbose=0)
    y_train = get_y(y_train, prediction_type)
    grid.fit(X_train_combined, y_train)
    rmtree(cachedir)
    
    scores = get_scores(prediction_type, grid, X_test_combined,
                        y_test, X_test_2, y_test_2)
    res = set_up_results(data, mode, run_config, prediction_type, fs_method,
                         k, estimator, repeat, scores, grid)
    return res


def get_y(y_df, prediction_type):
    """
    convert data frame to structured array for survival type
    """
    if prediction_type == 'sur':
        col_event = y_df.columns[0]
        col_time = y_df.columns[1]
        y = np.empty(dtype=[(col_event, np.bool), (col_time, np.float64)],
                    shape=y_df.shape[0])
        y[col_event] = (y_df[col_event] == 1).values
        y[col_time] = y_df[col_time].values
    else:
        y = y_df.values.ravel()
    return y


def run_single_estimator(data, run_config, k, repeat, fs_method,
                         estimator, mode, seed):
    n_view = len(data['desc']['view_names'])
    res = []
    if fs_method is None:
        # proms and proms_mo (if there are more than 1 view available)
        cur_res = run_single_fs_method(data, 'proms', run_config, k, repeat,
                      estimator, mode, seed)
    
        res.append(cur_res)
        if n_view > 1:
            cur_res = run_single_fs_method(data, 'proms_mo', run_config, k,
                          repeat, estimator, mode, seed)
            res.append(cur_res)
        if run_config['include_pca']:
            cur_res = run_single_fs_method(data, 'pca_ex', run_config, k,
                          repeat, estimator, mode, seed)
            res.append(cur_res)
    else:
        cur_res = run_single_fs_method(data, fs_method, run_config, k, repeat,
                      estimator, mode, seed)
        res.append(cur_res)

    return res


def prepare_data(all_data, repeat, mode):
    test_ratio = 0.3
    y = all_data['y']
    prediction_type = all_data['desc']['prediction_type']
    n_view = all_data['desc']['n_view']
    dataset_name = all_data['desc']['name']
    prediction_type = all_data['desc']['prediction_type']
    view_names = all_data['desc']['view_names']
    target_view_name = all_data['desc']['target_view']
    if mode == 'eval':
        if prediction_type == config.prediction_map['classification']:
            y_train, y_test = train_test_split(y, test_size=test_ratio,
                                  stratify=y, random_state=repeat)
        else:
            y_train, y_test = train_test_split(y, test_size=test_ratio,
                                  random_state=repeat)
    else:  # full model
        y_train, y_test = y, None

    data = {'desc': {},
            'train': {},
            'test': None,
            'test_2': None
            }
    data['desc']['name'] = dataset_name
    data['desc']['view_names'] = view_names
    data['desc']['target_view_name'] = target_view_name
    data['desc']['prediction_type'] = prediction_type
    if mode == 'eval':
        data['test'] = {}
    if all_data['X_test'] is not None:
        data['test_2'] = {}

    for i in range(n_view):
        cur_view_name = view_names[i]
        data['train'][cur_view_name] = {}
        if mode == 'eval':
            data['test'][cur_view_name] = {}
        cur_X = all_data['X'][cur_view_name]
        cur_X_train = cur_X.loc[y_train.index, :]
        data['train'][cur_view_name]['X'] = cur_X_train
        if mode == 'eval':
            cur_X_test = cur_X.loc[y_test.index, :]
            data['test'][cur_view_name]['X'] = cur_X_test

        if all_data['X_test'] is not None:
            data['test_2']['X'] = all_data['X_test']
            data['test_2']['y'] = all_data['y_test']

    data['train']['y'] = y_train
    if mode == 'eval':
        data['test']['y'] = y_test

    return data


def run_single_repeat(all_data, run_config, k, repeat, fs_method,
                      mode, seed):
    data = prepare_data(all_data, repeat, mode)
    estimators = run_config['estimators']
    res = []
    for estimator in estimators:
        cur_res = run_single_estimator(data, run_config, k, repeat, fs_method,
                      estimator, mode, seed)
        res.extend(cur_res)
    return res


def run_single_k(all_data, run_config, k, fs_method, mode, seed):
    n_repeats = run_config['repeat']
    res = []
    for repeat in range(n_repeats):
        cur_res = run_single_repeat(all_data, run_config, k, repeat,
                      fs_method, mode, seed)
        res.extend(cur_res)
    return res


def get_estimator_parameter_grid(estimator, prediction_type='cls'):
    """
    get parameter grid for pipeline
    """
    pg = config.parameter_grid[prediction_type]
    if not estimator in pg:
        raise ValueError(f'estimator "{estimator}" not supported for prediction type "{prediction_type}"')
    pipeline_pg = {}
    for parameter in pg[estimator]:
        pipeline_pg[estimator + '__' + parameter] = pg[estimator][parameter]

    return pipeline_pg


def get_results_col_name(all_data, mode='eval'):
    """
    set result data frame column names
    """
    prediction_type = all_data['desc']['prediction_type']
    check_prediction_type_ok(prediction_type)

    if mode == 'eval':
        if prediction_type == 'cls':
            column_names = ['fs', 'type', 'k', 'estimator', 'repeat',
                            'val_score', 'val_pred_label',
                            'val_label', 'val_acc', 'val_auroc']
        elif prediction_type == 'reg':
            column_names = ['fs', 'type', 'k', 'estimator', 'repeat',
                            'val_pred_label', 'val_label', 'val_mse', 'val_r2']
        elif prediction_type == 'sur':
            column_names = ['fs', 'type', 'k', 'estimator', 'repeat',
                            'val_label', 'val_risk_score', 'val_c_index']
    elif mode == 'full':
        if all_data['desc']['has_test']:
            if all_data['y_test'] is not None:
                if prediction_type == 'cls':
                    column_names = ['fs', 'type', 'k', 'estimator',
                                    'features', 'membership',
                                    'mean_val_acc', 'mean_val_auroc',
                                    'test_score', 'test_pred_label',
                                    'test_label', 'test_accuracy', 'test_auroc']
                elif prediction_type == 'reg':
                    column_names = ['fs', 'type', 'k', 'estimator',
                                    'features', 'membership',
                                    'mean_val_mse', 'mean_val_r2',
                                    'test_pred_label', 'test_label',
                                    'test_mse', 'test_r2']
                elif prediction_type == 'sur':
                    column_names = ['fs', 'type', 'k', 'estimator',
                                    'features', 'membership',
                                    'mean_val_c_index',
                                    'test_label', 'test_risk_score', 'test_c_index']
            else:
                if prediction_type == 'cls':
                    column_names = ['fs', 'type', 'k', 'estimator',
                                    'features', 'membership',
                                    'mean_val_acc', 'mean_val_auroc',
                                    'test_score', 'test_pred_label']
                elif prediction_type == 'reg':
                    column_names = ['fs', 'type', 'k', 'estimator',
                                    'features', 'membership',
                                    'mean_val_mse', 'mean_val_r2',
                                    'test_pred_label']
                elif prediction_type == 'sur':
                    column_names = ['fs', 'type', 'k', 'estimator',
                                    'features', 'membership',
                                    'mean_val_c_index', 'test_risk_score']
        else:
            if prediction_type == 'cls':
                column_names = ['fs', 'type', 'k', 'estimator',
                                'features', 'membership',
                                'mean_val_acc', 'mean_val_auroc']
            elif prediction_type == 'reg':
                column_names = ['fs', 'type', 'k', 'estimator',
                                'features', 'membership',
                                'mean_val_mse', 'mean_val_r2']
            if prediction_type == 'sur':
                column_names = ['fs', 'type', 'k', 'estimator',
                                'features', 'membership',
                                'mean_val_c_index']

    return column_names


def select_for_full_model(df, prediction_type):
    """
    select best configuration to train a full model
    """
    res = {}
    if prediction_type == 'cls':
        df_sel = df[['fs', 'k', 'estimator', 'val_acc', 'val_auroc']]
        # pca_ex cannot be used for full model
        df_sel = df_sel[df_sel.fs != 'pca_ex']
        df_sel = df_sel.groupby(['fs', 'k', 'estimator']).mean()
        df_sel = df_sel.reset_index()
        best_auroc_idx = df_sel['val_auroc'].argmax()
        fs_sel = df_sel.loc[best_auroc_idx, 'fs']
        k_sel = df_sel.loc[best_auroc_idx, 'k']
        estimator_sel = df_sel.loc[best_auroc_idx, 'estimator']
        best_mean_acc = df_sel.loc[best_auroc_idx, 'val_acc']
        best_mean_auroc = df_sel.loc[best_auroc_idx, 'val_auroc']
        res['fs_sel'] = fs_sel
        res['k_sel'] = k_sel
        res['estimator_sel'] = estimator_sel
        res['best_mean_acc'] = best_mean_acc
        res['best_mean_auroc'] = best_mean_auroc
    elif prediction_type == 'reg':
        df_sel = df[['fs', 'k', 'estimator', 'val_mse', 'val_r2']]
        df_sel = df_sel[df_sel.fs != 'pca_ex']
        df_sel = df_sel.groupby(['fs', 'k', 'estimator']).mean()
        df_sel = df_sel.reset_index()
        best_auroc_idx = df_sel['val_r2'].argmax()
        fs_sel = df_sel.loc[best_auroc_idx, 'fs']
        k_sel = df_sel.loc[best_auroc_idx, 'k']
        estimator_sel = df_sel.loc[best_auroc_idx, 'estimator']
        best_mean_mse = df_sel.loc[best_auroc_idx, 'val_mse']
        best_mean_r2 = df_sel.loc[best_auroc_idx, 'val_r2']
        res['fs_sel'] = fs_sel
        res['k_sel'] = k_sel
        res['estimator_sel'] = estimator_sel
        res['best_mean_mse'] = best_mean_mse
        res['best_mean_r2'] = best_mean_r2
    elif prediction_type == 'sur':
        df_sel = df[['fs', 'k', 'estimator', 'val_c_index']]
        df_sel = df_sel[df_sel.fs != 'pca_ex']
        df_sel = df_sel.groupby(['fs', 'k', 'estimator']).mean()
        df_sel = df_sel.reset_index()
        best_auroc_idx = df_sel['val_c_index'].argmax()
        fs_sel = df_sel.loc[best_auroc_idx, 'fs']
        k_sel = df_sel.loc[best_auroc_idx, 'k']
        estimator_sel = df_sel.loc[best_auroc_idx, 'estimator']
        best_mean_c_index = df_sel.loc[best_auroc_idx, 'val_c_index']
        res['fs_sel'] = fs_sel
        res['k_sel'] = k_sel
        res['estimator_sel'] = estimator_sel
        res['best_mean_c_index'] = best_mean_c_index

    return res


def run_fs(all_data, run_config, run_version, output_root, seed):
    k = run_config['k']
    dataset_name = all_data['desc']['name']
    prediction_type = all_data['desc']['prediction_type']

    # evaluate: performance evaluation, select features with
    #           model built with train set and evaluate in validation set
    column_names = get_results_col_name(all_data, 'eval')
    out_dir_run = os.path.join(output_root, run_version)
    res = []
    for cur_k in k:
        cur_res = run_single_k(all_data, run_config, cur_k, None, 'eval', seed)
        res.extend(cur_res)

    results_df = pd.DataFrame(res, columns=column_names)
    out_file = dataset_name + '_results_'
    out_file = out_file + run_version + '_eval.tsv'
    out_file = os.path.join(out_dir_run, out_file)
    results_df.to_csv(out_file, header=True, sep='\t', index=False)

    # we will select the best combination of fs, k, estimator
    # based on cross validation results (average score) from the previous step
    # to fit a full model
    res_dict = select_for_full_model(results_df, prediction_type)

    # # full: build full model with all data and test in another
    # #       indpendent data set if available
    column_names = get_results_col_name(all_data, 'full')
    # re-define run_config file
    run_config['repeat'] = 1
    k_sel = res_dict['k_sel']
    run_config['k'] = [k_sel]
    run_config['estimators'] = [res_dict['estimator_sel']]
    fs_method = res_dict['fs_sel']

    res = []
    cur_res = run_single_k(all_data, run_config, k_sel, fs_method,
                           'full', seed)
    if prediction_type == 'cls':
        cur_res[0][6] = res_dict['best_mean_acc']
        cur_res[0][7] = res_dict['best_mean_auroc']
    elif prediction_type == 'reg':
        cur_res[0][6] = res_dict['best_mean_mse']
        cur_res[0][7] = res_dict['best_mean_r2']
    elif prediction_type == 'sur':
        cur_res[0][6] = res_dict['best_mean_c_index']

    res.extend(cur_res)
    results_df = pd.DataFrame(res, columns=column_names)
    out_file = dataset_name + '_results_'
    out_file = out_file + run_version + '_full.tsv'
    out_file = os.path.join(out_dir_run, out_file)
    results_df.to_csv(out_file, header=True, sep='\t', index=False,
                      quoting=csv.QUOTE_NONE)


def check_data_config(config_file):
    """
    verify data configuration file
    """
    with open(config_file) as config_fh:
        data_config = yaml.load(config_fh, Loader=yaml.FullLoader)

    required_fields = {'project_name', 'data_directory', 'train_data_directory', 'target_view',
                       'target_label', 'data'}
    allowed_fields = required_fields | {'test_data_directory'}
    if not required_fields <= data_config.keys() <= allowed_fields:
        raise Exception(f'provided fields: {sorted(data_config.keys())}\n'
                        f'config required fields: {sorted(required_fields)}\n'
                        f'allowed fields: {sorted(allowed_fields)}')

    test_dataset = data_config['test_data_directory'] if 'test_data_directory' in data_config else None
    data_required_fields = {'train'}
    if test_dataset is not None:
        data_allowed_fields = {'test'} | data_required_fields
    else:
        data_allowed_fields = data_required_fields
    data_provided_fields = data_config['data'].keys()
    if not data_required_fields <= data_provided_fields <= data_allowed_fields:
        raise Exception(f'data section provided fields: {sorted(data_provided_fields)}\n'
                        f'required fields: {sorted(data_required_fields)}\n'
                        f'allowed fileds: {sorted(data_allowed_fields)}')

    train_required_fields = {'label', 'view'}
    train_provided_fields = data_config['data']['train'].keys()
    if not train_required_fields <= train_provided_fields:
        raise Exception(f'train data required fields: {sorted(train_provided_fields)}\n'
                        f'required fields: {sorted(train_required_fields)}')


def create_dataset(config_file, output_run):
    """ create data structure from input data files """
    print(f'data config file: {config_file}')
    check_data_config(config_file)
    with open(config_file) as config_fh:
        data_config = yaml.load(config_fh, Loader=yaml.FullLoader)
        data_root = data_config['data_directory']
        if not os.path.isabs(data_root):
            # relative to the data config file
            config_root = os.path.abspath(os.path.dirname(config_file))
            data_root = os.path.join(config_root, data_root)
        ds_name = data_config['project_name']
    all_data = Dataset(name=ds_name, root=data_root, config_file=config_file,
                       output_dir=output_run)()
    return all_data


def check_run_config(run_config_file, n_train_sample, prediction_type):
    with open(run_config_file) as config_fh:
        run_config = yaml.load(config_fh, Loader=yaml.FullLoader)
        if not 'n_jobs' in run_config:
            # assume running on a node with 4 cpus
            run_config['n_jobs'] = 4
        if not 'repeat' in run_config:
            run_config['repeat'] = 5
        if not 'k' in run_config:
            raise Exception('must specifiy k in run configuration file.')
        k = run_config['k']
        k_max = sorted(k)[-1]
        if k_max > int(0.25*n_train_sample):
            raise Exception('largest k should be less than 25% '
                    'of the number of training samples')
        default_estimators = config.default_estimators[prediction_type]             
        if not 'estimators' in run_config:
            run_config['estimators'] = default_estimators
        else:
            if not set(run_config['estimators']).issubset(default_estimators):
                raise Exception(f'supported estimators:{default_estimators}')
        if not 'percentile' in run_config:
            run_config['percentile'] = [1, 5, 10, 20]
        else:
            all_ok = all(x > 0 and x < 50 for x in run_config['percentile']) 
            if not all_ok: 
                raise Exception('all percentile values must be > 0 and < 50')
    
    return run_config


def main():
    # ignore warnings from joblib
    warnings.filterwarnings('ignore')
    os.environ['PYTHONWARNINGS'] = 'ignore'
    parser = get_parser()
    args = parser.parse_args()
    run_config_file = args.run_config_file
    data_config_file = args.data_config_file
    output_root = args.output_root
    include_pca = args.include_pca
    # time stamp
    run_version = args.run_version
    # random seed for full model (if applicable)
    seed = args.seed
    # prepare output directory
    out_dir_run = os.path.join(output_root, run_version)
    out_dir_run_full = os.path.join(out_dir_run, 'full_model')
    if not os.path.exists(out_dir_run_full):
        os.makedirs(out_dir_run_full)
    all_data = create_dataset(data_config_file, out_dir_run)
    n_train_sample = all_data['y'].shape[0]
    prediction_type = all_data['desc']['prediction_type']
    run_config = check_run_config(run_config_file, n_train_sample,
                     prediction_type)
    run_config['run_version'] = run_version
    run_config['output_root'] = output_root
    run_config['include_pca'] = include_pca

    run_fs(all_data, run_config, run_version, output_root, seed)


def is_valid_file(arg):
    """ check if the file exists """
    arg = os.path.abspath(arg)
    if not os.path.exists(arg):
        msg = "The file %s does not exist!" % arg
        raise argparse.ArgumentTypeError(msg)
    else:
        return arg


def date_time_now():
    """ get the current date and time """
    now = datetime.now()
    date_time = now.strftime("%Y_%m_%d_%H_%M_%S")
    return date_time


def get_parser():
    ''' get arguments '''
    parser = argparse.ArgumentParser()
    parser.add_argument('-f', '--file', dest='run_config_file',
                        type=is_valid_file,
                        required=True,
                        help='configuration file for the run',
                        metavar='FILE',
                        )
    parser.add_argument('-d', '--data', dest='data_config_file',
                        type=is_valid_file,
                        required=True,
                        help='configuration file for data set',
                        metavar='FILE',
                        )
    parser.add_argument('-s', '--seed', dest='seed',
                        default=42,
                        type=int,
                        help='random seed '
                        )
    parser.add_argument('-o', '--output', dest='output_root',
                        default='results',
                        type=str,
                        help='output directory'
                        )
    parser.add_argument('-r', '--run_version', dest='run_version',
                        default=date_time_now(),
                        type=str,
                        help='name of the run, default to current date/time'
                        )
    parser.add_argument('-p', '--include_pca', dest='include_pca',
                        default=False,
                        action='store_true',
                        help='include supervised PCA method in the results'
                       )
    return parser


if __name__ == '__main__':
    multiprocessing.set_start_method('forkserver')
    main()
